from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint.enlisting import Enlist
from pyadjoint import *
from pyadjoint_utils import *
from pyadjoint_utils.tensorflow_adjoint import (
    get_params_feed_dict,
    run_tensorflow_graph,
)
from pyadjoint_utils.fenics_adjoint import function_get_local, function_set_local
from .space_builders import DirectSum
from .types import PointMap
from .numpy import Ndarrays
import numpy as np
from crikit.fe import Function, project, grad


class CRTensorFlow(PointMap):
    def __init__(
        self,
        inputs,
        outputs,
        session,
        input_space=None,
        output_space=None,
        tf_kwargs=None,
    ):
        self.g_inputs = Enlist(inputs)
        self.g_outputs = Enlist(outputs)
        self.tf_kwargs = tf_kwargs if tf_kwargs is not None else {}
        self.session = session
        self.graph = self.session.graph

        with self.graph.as_default():
            self.feed_dict = get_params_feed_dict(self.session)
        self.params_keys = [g_p for g_p in self.feed_dict]
        self.params = [p for g_p, p in self.feed_dict.items()]
        self.params_controls = [Control(p) for p in self.params]

        if input_space is None:
            input_spaces = []
            for t in self.g_inputs:
                shape = tuple(-1 if s.value is None else s.value for s in t.shape)
                # Don't pass the dtype since it can break valid Space conversions.
                # We could fix the problem by adding an NdarrayToNdarray covering that handles dtype stuff.
                # input_spaces.append(Ndarrays(shape, dtype=t.dtype.as_numpy_dtype))
                input_spaces.append(Ndarrays(shape))
            input_space = DirectSum(input_spaces)
            input_space = self.g_inputs.delist(input_space)

        if output_space is None:
            output_spaces = []
            for t in self.g_outputs:
                shape = tuple(-1 if s.value is None else s.value for s in t.shape)
                # output_spaces.append(Ndarrays(shape, dtype=t.dtype.as_numpy_dtype))
                output_spaces.append(Ndarrays(shape))
            output_space = DirectSum(output_spaces)
            output_space = self.g_outputs.delist(output_space)

        super().__init__(input_space, output_space)

    def _set_feed_dict(self, inputs):
        assert len(inputs) == len(self.g_inputs), "%d != %d" % (
            len(inputs),
            len(self.g_inputs),
        )

        for input_val, g_input in zip(inputs, self.g_inputs):
            self.feed_dict[g_input] = input_val

    def __call__(self, args):
        self.inputs = Enlist(args)
        self._set_feed_dict(self.inputs)
        with self.session.as_default():
            outputs = run_tensorflow_graph(
                self.g_outputs, self.feed_dict, **self.tf_kwargs
            )
            outputs = self.g_outputs.delist(outputs)
        self.controls = [Control(a) for a in self.inputs]
        self.rf = ReducedFunction(outputs, self.controls)
        return outputs

    def numpy_call(self, args):
        args = Enlist(args)
        self._set_feed_dict(args)
        with self.session.as_default():
            outputs = run_tensorflow_graph(
                self.g_outputs, self.feed_dict, annotate=False
            )
            outputs = self.g_outputs.delist(outputs)
        return outputs

    def setParams(self, params):
        self.params = params
        self.params_controls = [Control(p) for p in self.params]

        for g_p, p in zip(self.params_keys, self.params):
            self.feed_dict[g_p] = p



import jax.numpy as jnp
from jax import nn
import numpy as np
import jax
from ..invariants import (
    InvariantInfo,
    TensorType,
    LeviCivitaType,
    get_invariant_functions,
    get_invariant_descriptions,
)
from jax.tree_util import (
    Partial as partial,
)  # JAX-friendlier version of functools.partial
from .map_builders import Callable as CallablePointMap
from .types import PointMap
from .space_builders import DirectSum
from pyadjoint_utils.jax_adjoint import array, ndarray, overload_jax, get_default_dtype
from .jax_utils import JAXArrays, UFLExprSpace_To_JAX, JAX_UFLFunctionSpace_Covering
from crikit.logging import logger
from typing import Union, Any, Optional, Tuple, Sequence, Iterable, Callable
from functools import wraps

Array = Any


class CR(PointMap):
    """A Constitutive Relation that automatically generates scalar and form
    invariants with :mod:`crikit.invariants`. All you need to provide is a function
    to compute scalar values of the scalar invariants that can be multiplied
    against the form invariants to form an equivariant tensor function in
    accordance with the canonical representation of Wineman and Pipkin, who
    showed that any equivariant (under a physical group) tensor function can be
    represented as a linear combination of scalar functions of scalar invariants
    and form invariants. In other words, this is a function that takes in
    the scalar invariants at a point as a one-dimensional JAX array (as well as any
    parameters you specify), and returns a one-dimensional JAX array, with one element
    for each form invariant.

    You can use :func:`cr_function_shape` to determine how many scalar invariants
    your function will take in and how many scalar values your function will
    need to output.

    """

    def __init__(
        self,
        output_type: TensorType,
        input_types: Sequence[TensorType],
        cr_function: Optional[Callable] = None,
        params: Optional[Sequence[Array]] = None,
        cr_static_argnums: Optional[Sequence[int]] = None,
        vmap: bool = True,
        vmap_inner: Optional[bool] = None,
        nojit: bool = False,
        strain_energy: bool = False,
        compiled_jacobian: bool = True,
        **cr_jax_kwargs,
    ):
        """Constructor for CR

        :param output_type: a TensorType corresponding to the output. If you want a
            strain-energy CR (one which computes the stress
            as the partial derivative of a strain energy functional with respect to the first
            input, then pass :meth:`crikit.invariants.TensorType.make_scalar` as the output type
            (i.e. a scalar).
        :type output_type: TensorType
        :param input_types: a sequence of TensorTypes corresponding to the inputs
        :type input_types: Sequence[TensorType]
        :param cr_function: The function to evaluate.
        :type cr_function: Callable, optional
        :param params: the initial values of the parameters, default None
        :type params: Sequence[jnp.ndarray], optional
        :param cr_static_argnums: the static_argnums parameter for :func:`jax.jit` for your cr_function
        :type cr_static_argnums: Union[int,Iterable[int]], optional
        :param vmap:  should we :func:`jax.vmap` the CR invariant functions over the inputs? True if your CR is going
            to be given input values at multiple points on a mesh (i.e. if the input is one second-order tensor in
            3-d, and you plan to evaluate the CR at multiple points at once by stacking the inputs, you want
            this to be True), default True
        :type vmap: bool, optional
        :param vmap_inner:  should we :func:`jax.vmap` the inner function over the inputs? True if your inner CR is
            going to be given input values at multiple points on a mesh that are handled independently of each other
            This defaults to the value of the vmap parameter.
        :type vmap_inner: bool, optional
        :param nojit: if True, do NOT jit-compile the CR function, defaults to False
        :type nojit: bool, optional
        :param strain_energy: if True, implies that this CR has a strain energy function -- that is, ``cr_function`` is a scalar
            function that gives the strain energy as a function of a symmetric second-order input (and possibly other inputs), and the CR
            computes the stress as the derivative of strain energy with respect to the symmetric second-order input. Defaults to False
        :type strain_energy: bool, optional
        :param compiled_jacobian: if True, compiles the function that
            stacks the pointwise Jacobians in the backend Jacobian-computing
            function in :class:`JAXBlock`. Identical to the `compile_jacobian_stack`
            parameter to :func:`pyadjoint_utils.jax_adjoint.overload_jax`. Defaults to True
        :type compiled_jacobian: bool, optional
        :return: a CR object
        :rtype: CR

        """
        self._compiled_jacobian = compiled_jacobian
        self._out_type = output_type
        self._in_types = tuple(input_types)
        self._dims = max(
            tuple(max(x.shape) if len(x.shape) > 0 else 0 for x in self._in_types)
        )
        self._invariant_info = InvariantInfo(self._dims, self._in_types, self._out_type)
        self._strain_energy = strain_energy
        _, self._in_types = self._invariant_info.get_group_symbol(
            sanitize_input_types=True
        )
        self._num_inputs = len(self._in_types)

        ignore_warnings = self._out_type.order == 0
        self._scalar_invt_func, self._form_invt_func = get_invariant_functions(
            self._invariant_info, suppress_warning_print=ignore_warnings
        )
        if isinstance(cr_function, str):
            self.load_tensorflow_model(cr_function)
        else:
            self._f = cr_function

        self._params = params or []
        self._num_params = len(self._params)
        self._diff_argnums = tuple(range(self._num_inputs + self._num_params))

        self._num_scalar_invts = None
        self._form_invt_shape = (
            None  # shape of the result of a call to self.form_invt_func()
        )
        self._determine_invariants_shape()
        # cr function takes one array containing scalar invariants and some params
        self._f_num_inputs = 1 + self._num_params
        self._vmap_axes = (0,) + (None,) * self._num_params
        if self._f is None:
            vmap = False

        vmap_inner = vmap if vmap_inner is None else vmap_inner
        if vmap_inner:
            self._f = jax.vmap(self._f, in_axes=self._vmap_axes)

        if vmap:
            self._scalar_invt_func = jax.vmap(self._scalar_invt_func)
            self._form_invt_func = jax.vmap(self._form_invt_func)
            self._invariant_evaluator = jax.vmap(partial(jnp.tensordot, axes=1))
            batched_in_shapes = tuple(
                tuple([-1] + list(x.shape)) for x in self._in_types
            )
            batched_out_shape = tuple(
                [-1]
                + list(
                    self._out_type.shape
                    if not self._strain_energy
                    else (self._dims, self._dims)
                )
            )
        else:
            self._invariant_evaluator = partial(jnp.tensordot, axes=1)
            batched_in_shapes = tuple(x.shape for x in self._in_types)
            batched_out_shape = (
                self._out_type.shape
                if not self._strain_energy
                else (self._dims, self._dims)
            )

        self._nojit = nojit
        self._vmap_rest = vmap
        self._vmap_inner = vmap_inner
        self._static_argnums = cr_static_argnums

        # overload (the backend of) self.__call__() so we can differentiate it with JAX
        self._overloaded_call = self._make_overloaded_call()
        source = (
            DirectSum([JAXArrays(bs) for bs in batched_in_shapes])
            if len(batched_in_shapes) > 1
            else JAXArrays(batched_in_shapes[0])
        )
        target = JAXArrays(batched_out_shape)
        super().__init__(source, target)

    @staticmethod
    def from_arrays(
        example_output: jnp.ndarray,
        example_inputs: Iterable[jnp.ndarray],
        cr_function: Optional[Callable] = None,
        params: Optional[Iterable[Array]] = None,
        cr_static_argnums: Optional[Sequence[int]] = None,
        vmap: bool = True,
        **kwargs,
    ):
        """The preferred way to construct a crikit.cr.CR if you don't want to manually construct the :class:`crikit.invariants.TensorType` s corresponding
        to your input and outputs tensor types. Ensure that, if your material has
        a structural tensor, you include it in `example_inputs`
        For example, a plank of wood is frequently modeled as being
        transverse-isotropic, with the structural tensor being a vector field
        pointing in the direction of the grain. If you want the symmetry to not include
        flips--that is, a subset of hemitropy instead of isotropy--ensure that you
        pass the Levi-Civita tensor (eps_ij or eps_ijk, depending on how many spatial
        dimensions you're in) as an :code:`example_input`, but DO NOT pass it into :func:`CR.__call__()`.
        If you pass the Levi-Civita tensor as an :code:`example_input`, we will account for
        its presence in the inputs without you passing it in.

        :param example_output: an example of what the output of the CR might look like; if that's a symmetric rank-two tensor,
            then example_output should also be that (e.g. :code:`jnp.eye(number_of_spatial_dimensions)`), etc.
        :type example_output: jnp.ndarray
        :param example_inputs: an iterable of JAX arrays of the same shape and symmetry as the inputs to the CR function
        :type example_inputs: Sequence[Array]
        :type cr_function: Either the function to evaluate OR a directory containing a saved TensorFlow model
            to load.
        :type cr_function: Union[function,str], optional
        :param params: the initial values of the parameters, default None
        :type params: Iterable[jnp.ndarray], optional
        :param cr_static_argnums: the static_argnums parameter for :func:`jax.jit` for your cr_function
        :type cr_static_argnums: Union[int,Iterable[int]], optional
        :param vmap:  should we :func:`jax.vmap` the CR function over the inputs? True if your CR is going to be
            given input values at multiple points on a mesh (i.e. if the input is one second-order tensor in
            3-d, and you plan to evaluate the CR at multiple points at once by stacking the inputs, you want
            this to be True), default True
        :type vmap: bool, optional
        :return: A crikit.cr.CR
        :rtype: CR

        """
        info = InvariantInfo.from_arrays(example_output, *example_inputs, **kwargs)
        cr = CR(
            info.output_type,
            info.input_types,
            cr_function=cr_function,
            params=params,
            static_argnums=cr_static_argnums,
            vmap=vmap,
            **kwargs,
        )
        return cr

    def __call__(self, inputs, **kwargs) -> Union[ndarray, Tuple[ndarray]]:
        """Evaluates the CR

        :param inputs: the inputs to the CR, as JAX arrays, or :class:`pyadjoint_utils.jax_adjoint.ndarray` s
           (if you're differentiating with Pyadjoint)
        :type inputs: Union[Iterable[pyadjoint_utils.jax_adjoint.ndarray,jnp.ndarray]]
        :return: The value of the invariant CR function (self.function) evaluated
           with the scalar and form-invariants generated by `inputs`
        :rtype: Union[ndarray, Tuple[ndarray]]

        """
        try:
            if not isinstance(self.source, DirectSum):
                inputs = (inputs,)
            self._check_inputs(inputs)
            params = kwargs.get("params", self._params)
            val = self._overloaded_call(*inputs, *params)
            return val
        except Exception as e:
            print(
                f"""Caught exception in CR.__call__():
            {e}
            Recall that inputs to CR.__call__() are passed directly to the scalar and form-invariant calculating functions."""
            )
            raise e

    def set_params(self, new_params: Iterable[Array]) -> None:
        if len(new_params) != len(self._params):
            raise ValueError(
                f"Tried to set_params() with {len(new_params)} params, but this CR expects {len(self._params)}!"
            )
        self._params = new_params

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate_scalar_jax_cr(self, *args):
        inputs = args[: self._num_inputs]
        params = args[self._num_inputs :]
        scalar_invariants = self._scalar_invt_func(*inputs)
        return self._f(scalar_invariants, *params)

    def _nojit_evaluate_scalar_jax_cr(self, *args):
        inputs = args[: self._num_inputs]
        params = args[self._num_inputs :]
        scalar_invariants = self._scalar_invt_func(*inputs)
        return self._f(scalar_invariants, *params)

    @partial(jax.jit, static_argnums=(0,))
    def _evaluate_jax_cr(self, *args):
        inputs = args[: self._num_inputs]
        params = args[self._num_inputs :]
        scalar_invariants = self._scalar_invt_func(*inputs)
        form_invariants = self._form_invt_func(*inputs)
        scalar_function_values = self._f(scalar_invariants, *params)
        # the generalized Wineman-Pipkin theorem of Zheng and Boehler (1994) guarantees that an extremely broad class of tensor functions
        # -- broad enough that we can say that, for our purposes, all tensor functions -- with a physical symmetry can be represented
        # as a linear combination of scalar functions of the scalar invariants, and form invariant functions
        return self._invariant_evaluator(scalar_function_values, form_invariants)

    def _nojit_evaluate_jax_cr(self, *args):
        inputs = args[: self._num_inputs]
        params = args[self._num_inputs :]
        scalar_invariants = self._scalar_invt_func(*inputs)
        form_invariants = self._form_invt_func(*inputs)
        scalar_function_values = self._f(scalar_invariants, *params)
        # the generalized Wineman-Pipkin theorem of Zheng and Boehler (1994) guarantees that an extremely broad class of tensor functions
        # -- broad enough that we can say that, for our purposes, all tensor functions -- with a physical symmetry can be represented
        # as a linear combination of scalar functions of the scalar invariants, and form invariant functions
        return self._invariant_evaluator(scalar_function_values, form_invariants)

    def load_tensorflow_model(self, directory: str) -> None:
        raise NotImplementedError  # not working right now
        import tensorflow as tf

        restored = tf.saved_model.load(directory)

        def model(invts, *params):
            return restored.f(
                tuple(tf.convert_to_tensor(x) for x in params),
                tf.convert_to_tensor(invts),
            )

        self._f = model

    def _make_overloaded_call(self):
        vmap = self._vmap_rest and self._vmap_inner
        pointwise = (vmap,) * self._num_inputs + (False,) * self._num_params
        if self._nojit:
            if self._strain_energy:

                def g(*args):
                    outs, pullback = jax.vjp(self._evaluate_scalar_jax_cr, *args)
                    # This assumes the strain is the first input.
                    return pullback(jnp.ones_like(outs))[0]

                return overload_jax(
                    g,
                    function_name=(self._f.__name__ if self._f else "JAX CR"),
                    argnums=self._diff_argnums,
                    static_argnums=self._static_argnums,
                    nojit=self._nojit,
                    pointwise=pointwise,
                    compile_jacobian_stack=self._compiled_jacobian,
                )
            f = (
                self._nojit_evaluate_jax_cr
                if self._out_type.order > 0
                else self._nojit_evaluate_scalar_jax_cr
            )
            return overload_jax(
                f,
                function_name=(self._f.__name__ if self._f else "JAX CR"),
                argnums=self._diff_argnums,
                static_argnums=self._static_argnums,
                nojit=self._nojit,
                pointwise=pointwise,
                compile_jacobian_stack=self._compiled_jacobian,
            )
        else:
            if self._strain_energy:

                def g(*args):
                    outs, pullback = jax.vjp(self._evaluate_scalar_jax_cr, *args)
                    # This assumes the strain is the first input.
                    return pullback(jnp.ones_like(outs))[0]

                return overload_jax(
                    g,
                    function_name=(self._f.__name__ if self._f else "JAX CR"),
                    argnums=self._diff_argnums,
                    static_argnums=self._static_argnums,
                    nojit=self._nojit,
                    pointwise=pointwise,
                    compile_jacobian_stack=self._compiled_jacobian,
                )
            # tuple(i+1 for i in range(len(self._params))),
            f = (
                self._evaluate_jax_cr
                if self._out_type.order > 0
                else self._evaluate_scalar_jax_cr
            )
            return overload_jax(
                f,
                function_name=(self._f.__name__ if self._f else "JAX CR"),
                argnums=self._diff_argnums,
                static_argnums=self._static_argnums,
                nojit=self._nojit,
                pointwise=pointwise,
                compile_jacobian_stack=self._compiled_jacobian,
            )

    @property
    def cr_input_shape(self):
        """
        The shape of the array of scalar invariants that the CR function takes
        as its first parameter.
        """
        if self._num_scalar_invts is None:
            self._determine_invariants_shape()

        return (self._num_scalar_invts,)

    @property
    def form_invariant_shape(self):
        """
        The shape of the array of form invariants
        """
        if self._form_invt_shape is None:
            self._determine_invariants_shape()

        return self._form_invt_shape

    @property
    def num_scalar_functions(self):
        """
        The number of scalar functions we need to make (each taking in the scalar invariants)
        in order to right-multiply the row vector of them against the form invariants
        For example, in 3d, an O(3)-invariant function of a symmetric rank-two tensor
        and a vector that outputs a symmetric rank-two has a _form_invt_shape of (6,3,3),
        so we need 6 scalar functions to make the right row vector to get a result of shape (3,3)
        """
        return self.form_invariant_shape[0] if len(self.form_invariant_shape) > 0 else 1

    def invariant_descriptions(
        self, ipython: Optional[bool] = None, html: Optional[bool] = None
    ) -> str:
        """
        A string describing both the scalar and form invariant functions, including their
        indices in the input/output of the CR.

        :param ipython: Are you in IPython mode? (e.g. in a Jupyter notebook) By default,
            tries to guess whether or not you are in IPython mode; set this manually if the
            behavior is not as desired.
        :type ipython: bool, optional
        :param html: Return an HTML string instead of a plain-text string? defaults to None, unless
            ``ipython`` is True, then True
        :return: A string describing the invariants
        :rtype: str
        """
        return get_invariant_descriptions(
            self._invariant_info, ipython=ipython, html=html
        )

    @property
    def function(self):
        return self._f

    @function.setter
    def function(self, f):
        """
        Set the CR function

        :param f: a function to set as the CR function
        :return: None
        """
        if isinstance(f, str):
            self.load_tensorflow_model(f)
        else:
            self._f = f
        self._overloaded_call = self._make_overloaded_call()

    def scalar_invariants(self, *inputs) -> jnp.ndarray:
        """
        Computes scalar invariants given inputs

        :param \\*inputs: the inputs to the CR
        :type \\*inputs: Iterable[jnp.ndarray]
        :return: A JAX DeviceArray containing the scalar invariants
        :rtype: jnp.ndarray
        """
        return self._scalar_invt_func(*inputs)

    def form_invariants(self, *inputs):
        """
        Computes form invariants given inputs

        :param \\*inputs: the inputs to the CR
        :type \\*inputs: Iterable[jnp.ndarray]
        :return: A JAX DeviceArray containing the stacked form-invariants
        :rtype: jnp.ndarray
        """
        return self._form_invt_func(*inputs)

    def save_model(self, directory):
        """Save the internal function (the one you can pass as ``cr_function`` in :meth:`CR.__init__` ) of a
        JAX-based CR to a directory by converting it to a TensorFlow model
        and then saving that. You can recover the model by using :meth:`CR.load_tensorflow_model`.

        :param directory: The directory name to save it to
        :type directory: str

        :return: None

         .. todo:: figure out a format in which we can save both the input/output types of the CR and the tf model together so we can reconstruct a whole CR object from a file and check that it matches this object's input and output types

        """
        save_jax_cr(self, directory)

    def _check_inputs(self, inputs):
        if len(inputs) != self._num_inputs:
            raise ValueError(
                f"Expected {self._num_inputs} inputs, but received {len(inputs)}!"
            )

        for i, (inpt, expected) in enumerate(zip(inputs, self._in_types)):
            shape = inpt.shape[1:] if self._vmap_rest else inpt.shape
            if tuple(shape) != tuple(expected.shape):
                if expected.shape == ():
                    if shape == (1,):
                        continue

                raise ValueError(
                    f"Expected input {i+1} to have shape {expected.shape}, but it has {shape}!"
                )

    def _determine_invariants_shape(self):
        _, in_types = self._invariant_info.get_group_symbol(sanitize_input_types=True)
        example_inputs = tuple(ipt.get_array_like() for ipt in in_types)

        self._num_scalar_invts = self._scalar_invt_func(*example_inputs).size
        self._form_invt_shape = (
            self._form_invt_func(*example_inputs).shape
            if self._out_type.order > 0
            else ()
        )

    def get_point_maps(self):
        """
        This method returns a :class:`PointMap` for each of the four functions used to the compute the CR output.

        The :meth:`CR.__call__` method takes ``inputs`` (and optionally ``params`` as a keyword arg)
        and uses four separate functions to compute the CR output:
            1. The scalar invariant function computes the scalar invariants as a function of ``inputs``.
            2. The form invariant function computes the form-invariant basis as a function of ``inputs``.
            3. The inner function computes the basis coefficients as a function of the scalar invariants and ``params``.
            4. The coefficient form function computes the CR output using the basis coefficients and the form-invariant basis.
        """
        b = (-1,) if self._vmap_rest else ()
        scalar_invt_space = JAXArrays(b + self.cr_input_shape)
        form_invt_tensor_space = JAXArrays(b + self._form_invt_shape)
        form_invt_coeff_space = JAXArrays(b + (self.num_scalar_functions,))

        inner_func = overload_jax(
            self._f,
            static_argnums=self._static_argnums,
            nojit=self._nojit,
            pointwise=(self._vmap_inner,) + (False,) * self._num_params,
            compile_jacobian_stack=self._compiled_jacobian,
        )

        @wraps(self._f)
        def inner_func_default_params(*inputs, params=None):
            if params is None:
                params = self._params
            return inner_func(*inputs, *params)

        inner_map = CallablePointMap(
            scalar_invt_space,
            form_invt_coeff_space,
            inner_func_default_params,
            bare=False,
        )

        scalar_invt_func = overload_jax(
            self._scalar_invt_func,
            nojit=self._nojit,
            pointwise=self._vmap_rest,
            compile_jacobian_stack=self._compiled_jacobian,
        )
        scalar_invt_map = CallablePointMap(
            self.source,
            scalar_invt_space,
            scalar_invt_func,
            bare=len(self._in_types) > 1,
        )

        form_invt_func = overload_jax(
            self._form_invt_func,
            nojit=self._nojit,
            pointwise=self._vmap_rest,
            compile_jacobian_stack=self._compiled_jacobian,
        )
        form_invt_map = CallablePointMap(
            self.source,
            form_invt_tensor_space,
            form_invt_func,
            bare=len(self._in_types) > 1,
        )

        coeff_form_func = overload_jax(
            self._invariant_evaluator,
            nojit=self._nojit,
            pointwise=self._vmap_rest,
            compile_jacobian_stack=self._compiled_jacobian,
        )
        coeff_form_map = CallablePointMap(
            DirectSum(scalar_invt_space, form_invt_coeff_space),
            self.target,
            coeff_form_func,
            bare=True,
        )

        return scalar_invt_map, form_invt_map, inner_map, coeff_form_map


def cr_function_shape(
    output: Union[Array, TensorType],
    inputs: Union[Sequence[TensorType], Sequence[Array]],
) -> Tuple[int, int]:
    """
    Computes the number of scalar invariants that a CR function for
    given inputs and outputs must take, as well as the number of scalar values
    that function must output to generate an invariant CR, and returns a tuple
    of (num_scalar_invariants,num_output_scalar_values).

    :param output: either an array (Numpy or JAX) or a TensorType representing the
       correct shape and symmetry of an output tensor from this CR.
    :type output: Union[Array,TensorType]
    :param inputs: an Iterable of either TensorType instances or arrays of the correct
       shape and symmetry as the input tensors of this CR; must contain the same
       type as output (i.e. if output is a TensorType, inputs must contain only
       TensorTypes, and likewise if output is an array, inputs must only contain
       arrays.
    :type inputs: Union[Sequence[TensorType], Sequence[Array]]
    :return: A tuple of (number of scalar invariants, number of output scalar values)
    :rtype: tuple
    """
    # NOTE: this definitely isn't the most efficient possible implementation, since constructing a CR does some other things that aren't
    # related to this computation, but this function is probably only called at most once per program execution during setup, and is
    # pretty cheap anyway
    if isinstance(output, TensorType):
        for ip in inputs:
            if not isinstance(ip, TensorType):
                raise TypeError(
                    f"If the output is a TensorType, all inputs must be TensorTypes, but you passed a {type(ip)}!"
                )
        cr = CR(output, inputs)

    array_types = (
        ndarray,
        jax.interpreters.xla.DeviceArray,
        jnp.ndarray,
        np.ndarray,
        float,
        int,
    )
    if isinstance(output, array_types):
        for ip in inputs:
            if not isinstance(ip, array_types):
                raise TypeError(
                    f"If the output is an array, all inputs must be arrays too, but you passed a {type(ip)}!"
                )
        cr = CR.from_arrays(output, inputs)

    num_scalar_invts = cr.cr_input_shape[0]
    num_scalar_funcs = cr.num_scalar_functions
    return (num_scalar_invts, num_scalar_funcs)


def save_jax_cr(cr: CR, directory: str):
    """Save a JAX-based CR to a directory by converting it to a TensorFlow model
    and then saving that. You can recover the model by using :meth:`CR.load_tensorflow_model`.

    :param cr: The CR to save
    :type cr: CR
    :param directory: The directory name to save it to
    :type directory: str

    :return: None

    """
    import tensorflow as tf
    from jax.experimental import jax2tf

    model = tf.Module()
    jax_dtype = get_default_dtype()
    tf_dtype = tf.float64 if jax_dtype == np.float64 else tf.dtype(jax_dtype)
    model.f = tf.function(
        jax2tf.convert(cr.function), autograph=False
    )  # , input_signature=[tf.TensorSpec(cr.cr_input_shape, tf_dtype)])

    tf.saved_model.save(model, directory)


class BlockCR(CR):
    """A vmapped-CR (i.e. a CR with ``vmap=True`` passed in the constructor, which is currently the default value)
    that passes the scalar invariants as one block to the function (i.e. without vmapping the function itself).
    """

    def __init__(
        self,
        output_type: TensorType,
        input_types: Sequence[TensorType],
        cr_function: Optional[Callable] = None,
        params: Optional[Sequence[Array]] = None,
        cr_static_argnums: Optional[Sequence[int]] = None,
        nojit: bool = False,
        strain_energy: bool = False,
        **cr_jax_kwargs,
    ):
        """

        :param output_type: a TensorType corresponding to the output. If you want a strain-energy CR (one which computes the stress
            as :math:`\sigma = \dfrac{\partial W}{\partial\varepsilon}` ), then pass :meth:`TensorType.make_scalar` as the output type
            (i.e. a scalar).
        :type output_type: TensorType
        :param input_types: a sequence of TensorTypes corresponding to the inputs
        :type input_types: Sequence[TensorType]
        :param cr_function: The function to evaluate.
        :type cr_function: Callable, optional
        :param params: the initial values of the parameters, default None
        :type params: Sequence[jnp.ndarray], optional
        :param cr_static_argnums: the static_argnums parameter for :func:`jax.jit` for your cr_function
        :type cr_static_argnums: Union[int,Iterable[int]], optional
        :param nojit: if True, do NOT jit-compile the CR function, defaults to False
        :type nojit: bool, optional
        :param strain_energy: if True, implies that this CR has a strain energy function -- that is, ``cr_function`` is a scalar
            function that gives the strain energy as a function of a symmetric second-order input (and possibly other inputs), and the CR
            computes the stress as the derivative of strain energy with respect to the symmetric second-order input. Defaults to False
        :type strain_energy: bool, optional
        :return: a BlockCR object
        :rtype: BlockCR

        """
        super().__init__(
            output_type,
            input_types,
            cr_function=cr_function,
            params=params,
            vmap=True,
            vmap_inner=False,
            cr_static_argnums=cr_static_argnums,
            nojit=nojit,
            strain_energy=strain_energy,
            **cr_jax_kwargs,
        )

    def _check_inputs(self, inputs):
        if len(inputs) != self._num_inputs:
            raise ValueError(
                f"Expected {self._num_inputs} inputs, but received {len(inputs)}!"
            )

        for i, (inpt, expected) in enumerate(zip(inputs, self._in_types)):
            shape = inpt.shape[1:]
            if tuple(shape) != tuple(expected.shape):
                if expected.shape == ():
                    if shape == (1,):
                        continue

                raise ValueError(
                    f"Expected input {i+1} to have shape {expected.shape}, but it has {shape}!"
                )

    @staticmethod
    def from_arrays(
        example_output: jnp.ndarray,
        example_inputs: Iterable[jnp.ndarray],
        cr_function: Optional[Callable] = None,
        params: Optional[Iterable[Array]] = None,
        cr_static_argnums: Optional[Sequence[int]] = None,
        vmap: bool = True,
        **kwargs,
    ):
        """The preferred way to construct a crikit.cr.CR if you don't want to manually construct the :class:`TensorType` s corresponding
        to your input and outputs tensor types. Ensure that, if your material has
        a structural tensor, you include it in `example_inputs`
        For example, a plank of wood is frequently modeled as being
        transverse-isotropic, with the structural tensor being a vector field
        pointing in the direction of the grain. If you want the symmetry to not include
        flips--that is, a subset of hemitropy instead of isotropy--ensure that you
        pass the Levi-Civita tensor (eps_ij or eps_ijk, depending on how many spatial
        dimensions you're in) as an ``example_input``, but DO NOT pass it into :func:`CR.__call__()`.
        If you pass the Levi-Civita tensor as an ``example_input``, we will account for
        its presence in the inputs without you passing it in.

        :param example_output: an example of what the output of the CR might look like; if that's a symmetric rank-two tensor,
            then example_output should also be that (e.g. :code:`jnp.eye(number_of_spatial_dimensions)`), etc.
        :type example_output: jnp.ndarray
        :param example_inputs: an iterable of JAX arrays of the same shape and symmetry as the inputs to the CR function
        :type example_inputs: Sequence[Array]
        :type cr_function: Either the function to evaluate OR a directory containing a saved TensorFlow model
            to load.
        :type cr_function: Union[function,str], optional
        :param params: the initial values of the parameters, default None
        :type params: Iterable[jnp.ndarray], optional
        :param cr_static_argnums: the static_argnums parameter for :func:`jax.jit` for your cr_function
        :type cr_static_argnums: Union[int,Iterable[int]], optional
        :return: A crikit.cr.BlockCR
        :rtype: CR
        """
        info = InvariantInfo.from_arrays(example_output, *example_inputs, **kwargs)
        cr = BlockCR(
            info.output_type,
            info.input_types,
            cr_function=cr_function,
            params=params,
            static_argnums=cr_static_argnums,
            **kwargs,
        )
        return cr


class P_Laplacian(CR):
    """A CR that represents a p-Laplacian."""

    def __init__(self, a, p, spatial_dims=3, eps2=1.0e-12, vmap=True):
        input_types = (TensorType.make_symmetric(2, spatial_dims),)
        output_type = TensorType.make_scalar()
        self.eps2 = eps2
        self.dims = spatial_dims
        cr_function = self._eval_p_laplacian
        super().__init__(
            output_type,
            input_types,
            cr_function,
            params=(
                a,
                p,
            ),
            vmap=vmap,
            nojit=False,
            strain_energy=False,
        )

    def _eval_p_laplacian(self, scalar_invts, a, p):
        rval = jnp.zeros((self.dims,))
        return jax.ops.index_update(
            rval, jax.ops.index[1], a * (scalar_invts[1] + self.eps2) ** ((p - 2) / 2)
        )


class RivlinModel(CR):
    """A CR that represents a Rivlin model -- that is, one of the form
    :math:`W = \sum\limits_{i=0}^n\sum\limits_{j=0}^n C_{ij} (I_1 - 3)^i (I_2 - 3)^j + \sum\limits_{k=1}^m D_k (J - 1)^{2k}`, where
    :math:`J = \mathrm{det}(B)`.
    """

    def __init__(self, C, D=None, spatial_dims=3, vmap=True, optimize_d=False):
        """

        :param C: The material constants :math:`C_{ij}`
        :type C: ndarray
        :param D: The material constants :math:`D_k`, defaults to None
        :type D: ndarray, optional
        :param spatial_dims: how many spatial dimensions? defaults to 3
        :type spatial_dims: int, optional
        :param vmap: the ``vmap`` parameter of :meth:`CR.__init__` , defaults to True
        :type vmap: bool, optional
        :param optimize_d: Controls which parameter we're optimizing the CR
            with respect to. If True, optimize ``D``, else optimize ``C``.
            Defaults to False.
        :type optimize_d: bool, optional
        :returns: a :class:`RivlinModel`
        :rtype: RivlinModel

        """

        self._C = C
        self._D = D
        self.incompressible = D is None
        self.n = C.shape[0]
        if C.shape[1] < C.shape[0]:
            raise ValueError("Must pass square array of parameter values C!")

        self.m = 0 if self.incompressible else D.size
        # inputs are just the left Cauchy-Green tensor
        input_types = (TensorType.make_symmetric(2, spatial_dims),)
        output_type = TensorType.make_scalar()
        cr_function = (
            self._eval_incompressible_rivlin_model
            if self.incompressible
            else self._eval_rivlin_model
        )
        if not self.incompressible:
            cr_function = (
                self._eval_rivlin_model_d if optimize_d else self._eval_rivlin_model
            )

        params = (D,) if optimize_d else (C,)
        super().__init__(
            output_type,
            input_types,
            cr_function,
            params=params,
            vmap=vmap,
            nojit=False,
            strain_energy=True,
        )
        self._params = params

    @property
    def C(self):
        return self._C

    @C.setter
    def set_C(self, new_C):
        self._params[0] = new_C
        self._C = new_C

    @property
    def D(self):
        return self._D

    @D.setter
    def set_D(self, new_D):
        self._params[1] = new_D
        self._D = new_D

    def _eval_rivlin_model(self, scalar_invts, C):
        I_1 = scalar_invts[0]
        I_2 = 0.5 * (scalar_invts[0] ** 2 - scalar_invts[1])
        J = self._det_from_traces(scalar_invts)

        W = np.array(0)
        for i in range(self.n):
            for j in range(self.n):
                W = W + C[i, j] * (I_1 - 3) ** i * (I_2 - 3) ** j

        for k in range(self.m):
            W = W + self._D[k] * (J - 1) ** (2 * k)

        return W

    def _eval_rivlin_model_d(self, scalar_invts, D):
        I_1 = scalar_invts[0]
        I_2 = 0.5 * (scalar_invts[0] ** 2 - scalar_invts[1])
        J = self._det_from_traces(scalar_invts)

        W = np.array(0)
        for i in range(self.n):
            for j in range(self.n):
                W = W + self._C[i, j] * (I_1 - 3) ** i * (I_2 - 3) ** j

        for k in range(self.m):
            W = W + D[k] * (J - 1) ** (2 * k)

        return W

    def _eval_incompressible_rivlin_model(self, scalar_invts, C):
        I_1 = scalar_invts[0]
        I_2 = 0.5 * (scalar_invts[0] ** 2 - scalar_invts[1])

        W = np.array(0)
        for i in range(self.n):
            for j in range(self.n):
                W = W + C[i, j] * (I_1 - 3) ** i * (I_2 - 3) ** j

        return W

    def _det_from_traces(self, scalar_invts):
        """
        The handy-dandy calculator at https://demonstrations.wolfram.com/TheDeterminantUsingTraces/
        tells us that the determinant of a 3-by-3 matrix :math:`A` is :math:`|A| = \frac{\text{tr}(A)^3}{6} - \frac{1}{2}\text{tr}(A^2)\text{tr}(A) + \frac{\text{tr}(A^3)}{3}`
        """
        ta, ta2, ta3 = scalar_invts[0], scalar_invts[1], scalar_invts[2]
        return (ta ** 3) / 6 - (ta2 * ta) / 2 + ta3 / 3



from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint.overloaded_type import create_overloaded_object
from .fe import *
from .fe_adjoint import *

import ufl
from fenics_adjoint.blocks.solving import GenericSolveBlock as SolveBlock

# Copied from fenics/dolfin/python/dolfin/fem/projection.py. Added dx parameter.
def project_backend(
    v,
    V,
    bcs=None,
    mesh=None,
    function=None,
    solver_type="lu",
    preconditioner_type="default",
    form_compiler_parameters=None,
    dx=None,
):
    # Note: I removed the beginning part handling when V is None or V is a MultiMeshFunctionSpace.

    # Ensure we have a mesh and attach to measure
    if mesh is None:
        mesh = V.mesh()

    ######### This is the only part I modified. ##############
    if dx is None:
        dx = ufl.dx(mesh)

        def get_dx(el, dx):
            if el.family() == "Quadrature":
                quad_params = {
                    "quadrature_rule": el.quadrature_scheme(),
                    "quadrature_degree": el.degree(),
                    "representation": "quadrature",
                }
                dx = dx(metadata=quad_params)
            return dx

        dx = get_dx(V.ufl_element(), dx)
        if hasattr(v, "function_space"):
            dx = get_dx(v.function_space().ufl_element(), dx)
    #########################################################

    # Define variational problem for projection
    w = TestFunction(V)
    Pv = TrialFunction(V)
    a = ufl.inner(w, Pv) * dx
    L = ufl.inner(w, v) * dx

    # Assemble linear system
    A, b = assemble_system(
        a, L, bcs=bcs, form_compiler_parameters=form_compiler_parameters
    )

    # Solve linear system for projection
    if function is None:
        function = Function(V)
    cpp.la.solve(A, function.vector(), b, solver_type, preconditioner_type)

    return function


# Directly copied from pyadjoint/fenics_adjoint/projection.py
def project(*args, **kwargs):
    """The project call performs an equation solve, and so it too must be annotated so that the
    adjoint and tangent linear models may be constructed automatically by pyadjoint.

    To disable the annotation of this function, just pass :py:data:`annotate=False`. This is useful in
    cases where the solve is known to be irrelevant or diagnostic for the purposes of the adjoint
    computation (such as projecting fields to other function spaces for the purposes of
    visualisation)."""

    annotate = annotate_tape(kwargs)
    with stop_annotating():
        output = project_backend(*args, **kwargs)
    output = create_overloaded_object(output)

    if annotate:
        bcs = kwargs.pop("bcs", [])
        dx = kwargs.pop("dx", None)
        block = ProjectBlock(args[0], args[1], output, bcs, dx)

        tape = get_working_tape()
        tape.add_block(block)

        block.add_output(output.block_variable)

    return output


# Directly copied from pyadjoint/fenics_adjoint/projection.py
class ProjectBlock(SolveBlock):
    def __init__(self, v, V, output, bcs=[], dx=None, *args, **kwargs):
        if dx is None:
            dx = ufl.dx(V.mesh())

        w = TestFunction(V)
        Pv = TrialFunction(V)
        a = inner(w, Pv) * dx
        L = inner(w, v) * dx

        super(ProjectBlock, self).__init__(a, L, output, bcs, *args, **kwargs)



from ..reduced_function import ReducedFunction
from pyadjoint.enlisting import Enlist
from functools import partial


def _default_apply_bc(bc, u):
    if hasattr(u, "vector"):
        bc.apply(u.vector())
    else:
        bc.apply(u)
    return u


def _is_valid_bc(bc):
    return hasattr(bc, "apply")


class ReducedEquation(object):
    """A class that encapsulates all the information required to formulate a
    reduced equation solve problem."""

    """
        reduced_function: a ReducedFunction that calculates the residual as a function of x.
        bcs: a function or list of functions that apply boundary conditions.
        h_bcs: a function or list of functions that apply homogenenous boundary conditions.
    """

    def __init__(self, reduced_function, bcs=None, h_bcs=None):
        bcs = [] if bcs is None else Enlist(bcs)
        h_bcs = [] if h_bcs is None else Enlist(h_bcs)

        bcs, h_bcs = self._validate_arguments(reduced_function, bcs, h_bcs)

        self.reduced_function = reduced_function
        self.bcs = bcs
        self.h_bcs = h_bcs

    def _validate_arguments(self, reduced_function, bcs, h_bcs):

        if not isinstance(reduced_function, ReducedFunction):
            raise TypeError("reduced_function should be a ReducedFunction")

        for i, bc in enumerate(bcs):
            if not callable(bc):
                if not _is_valid_bc(bc):
                    raise TypeError("Boundary conditions must have an 'apply' method!")
                else:
                    bcs[i] = partial(_default_apply_bc, bc)

        for i, h_bc in enumerate(h_bcs):
            if not callable(bc):
                if not _is_valid_bc(h_bc):
                    raise TypeError(
                        "Homogenized boundary conditions must have an 'apply' method!"
                    )
                else:
                    h_bcs[i] = partial(_default_apply_bc, h_bc)

        return bcs, h_bcs



from .tensorflow_block import get_params_feed_dict, run_tensorflow_graph

try:
    # Tensorflow 1.14 uses abseil, which breaks the logger.
    # https://github.com/tensorflow/tensorflow/issues/26691
    import absl.logging

    # https://github.com/abseil/abseil-py/issues/99
    absl.logging.root.removeHandler(absl.logging._absl_handler)
    # https://github.com/abseil/abseil-py/issues/102
    absl.logging._warn_preinit_stderr = False
except Exception:
    pass



from crikit.cr.space_builders import DirectSum
from crikit.cr.map_builders import CompositePointMap, IdentityPointMap, ParallelPointMap
from crikit.cr.types import Space, PointMap
from typing import Optional


class Covering:
    """Base class for Covering types.

    To be compatible with the covering params structure, the constructor
    should accept all keywords arguments and just ignore the ones that it
    doesn't need.

    """

    def __init__(self, base_space, covering_space, **covering_params):
        self._base_space = base_space
        self._covering_space = covering_space

    def covering_map(self, **params) -> PointMap:
        """This method must be overridden.

        Should return a map from the covering space to the base space.

        Args:
            **params: any parameters that should be used for creating the map.

        Returns:
            PointMap: A map from the covering space to the base space.

        """
        raise NotImplementedError

    def section_map(self, **params) -> PointMap:
        """This method must be overridden.

        Should return a map from the base space to the covering space.

        Args:
            **params: any parameters that should be used for creating the map.

        Returns:
            PointMap: A map from the base space to the covering space.

        """
        raise NotImplementedError

    def bundle_map(self, covering_params=None, section_params=None) -> PointMap:
        raise NotImplementedError


_covering_registry = {}
_default_covering_params = {}


def get_default_covering_params() -> dict:
    """
    Returns:
        dict: a reference to the current covering params dictionary
    """
    return _default_covering_params


def set_default_covering_params(*args, **kwargs) -> None:
    """Update the covering params dictionary.

    Args:
        *args: dictionaries with key-value pairs to add to the covering params
            dictionary.
        **kwargs: key-value pairs to add to the covering params dictionary.
    """
    for a in args:
        _default_covering_params.update(a)
    _default_covering_params.update(kwargs)


def reset_default_covering_params() -> None:
    """Resets the default covering params to an empty dictionary."""
    global _default_covering_params
    _default_covering_params = {}


def register_covering(
    base_space: Space, covering_space: Space, covering_class: Optional[Covering] = None
) -> Covering:
    """Register a covering class for use in :func:`get_map`

    The ``covering_class`` should be defined with a ``covering_map`` method that
    returns a point map mapping from ``covering_space`` to ``base_space``, and a
    ``section_map`` method that returns a point map from ``base_space`` to
    ``covering_space``.

    This function can be used as a class decorator, in which case the
    ``covering_class`` doesn't need to be specified.

    Args:
        base_space (type): a Space subclass.
        covering_space (type): a Space subclass.
        covering_class (type): the Covering subclass to register as handling
            (base_space, covering_space) mappings.

    Returns:
        type: returns ``covering_class`` such that it can be used as a decorator.

    """
    if covering_class is None:
        # decorator mode.
        def decorator(covering_class):
            register_covering(base_space, covering_space, covering_class)
            return covering_class

        return decorator
    _covering_registry[(base_space, covering_space)] = covering_class
    return covering_class


def get_map(source: Space, target: Space, **covering_params) -> PointMap:
    """Creates a map from ``source`` to ``target`` using the Covering registry. If a
    mapping from source to target is not found in the registry, it raises an
    exception.

    If ``source`` and ``target`` are DirectSums of multiple spaces, they must each
    have the same number of spaces. In that case, the i-th subspace in the
    source is mapped individually to the i-th subspace in the target space.

    Any additional kwargs are passed to the covering constructor.

    Args:
        source (Space): The source space.
        target (Space): The target space.
        **covering_params: Parameters to pass to the covering class constructor.

    Returns:
        PointMap: A map from the source space to the target space.

    """
    if source == target:
        return IdentityPointMap(s, t)

    if (
        isinstance(source, DirectSum)
        and isinstance(target, DirectSum)
        and len(source) == len(target)
    ):
        return ParallelPointMap(
            *tuple(get_map(s, t, **covering_params) for s, t in zip(source, target))
        )

    params = get_default_covering_params().copy()
    params.update(covering_params)

    s_type = type(source)
    t_type = type(target)
    covering_class = _covering_registry.get((s_type, t_type), None)
    if covering_class is not None:
        covering = covering_class(source, target, **params)
        return covering.section_map()
    covering_class = _covering_registry.get((t_type, s_type), None)
    if covering_class is not None:
        covering = covering_class(target, source, **params)
        return covering.covering_map()
    msg = "Don't have a covering for %s and %s with shapes %s and %s" % (
        s_type,
        t_type,
        source.shape(),
        target.shape(),
    )
    msg = msg + "\n        source space: %s, target space: %s" % (source, target)
    raise NotImplementedError(msg)


def get_composite_cr(*args, **covering_params) -> PointMap:
    """Creates a map that composes the given PointMaps and/or Spaces by using
    covering maps to convert between Spaces.

    Any kwargs are passed to :func:`get_map`.

    For example, ``get_composite_cr(space1, cr1, cr2, space2)`` returns a
    :class:`~crikit.cr.map_builders.CompositePointMap` that does the following:

    - takes input in ``space1``,
    - converts it to input to ``cr1``,
    - applies ``cr1``
    - converts the output to the input space of ``cr2``
    - applies ``cr2``
    - and then converts the output to ``space2``.

    The conversion point maps are created using the :func:`get_map` function.

    Args:
        *args (Space or PointMap): the PointMaps that should be applied and any
            desired spaces to convert to.
        **covering_params: Parameters to pass to the covering class constructor.

    Returns:
        PointMap: A map from the source space to the target space.

    """

    if len(args) == 0:
        return

    if isinstance(args[0], PointMap):
        previous_space = args[0].source
    elif isinstance(args[0], Space):
        previous_space = args[0]
    else:
        raise ValueError(
            "Found input argument that is neither a Space nor a PointMap:"
            + str(args[0])
        )

    maps = []
    for a in args:
        if isinstance(a, PointMap):
            # Map from previous space to this point map's input space and then apply the point map.
            if previous_space != a.source:
                maps.append(get_map(previous_space, a.source, **covering_params))
            maps.append(a)
            previous_space = a.target
        elif isinstance(a, Space):
            # Map from previous space to current space.
            if previous_space != a:
                maps.append(get_map(previous_space, a, **covering_params))
            previous_space = a
        else:
            raise ValueError(
                "Found input argument that is neither a Space nor a PointMap:" + str(a)
            )
    return CompositePointMap(maps)



import jax.numpy as jnp
import numpy as np
import jax
from jax.ops import index_update
from jax.tree_util import (
    register_pytree_node,
    register_pytree_node_class,
    tree_flatten,
    tree_unflatten,
)
from jax.tree_util import Partial as partial
from pyadjoint.overloaded_function import overload_function
from pyadjoint.overloaded_type import (
    OverloadedType,
    register_overloaded_type,
    create_overloaded_object,
)
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint_utils.block import Block
from pyadjoint_utils.convert import make_convert_block
from pyadjoint_utils.adjfloat import AdjFloat
from pyadjoint import AdjFloat, Control
from typing import Any, Iterable, Sequence, Union, Optional, Tuple
from .blocks import AddBlock, SubBlock, PowBlock, MulBlock, DivBlock, NegBlock

Array = Any
Shape = Sequence[int]

_default_dtype = jnp.float64


def set_default_dtype(dtype: jnp.dtype) -> None:
    """Sets the default data type for jax arrays used inside crikit. Default jax.numpy.float64.

    :param dtype: the default data type to set
    :type dtype: Union[jnp.dtype,str,Tuple[Union[jnp.dtype,str]]]
    :returns: None

    """
    global _default_dtype
    dtype = jnp.dtype(dtype)
    _default_dtype = dtype


def get_default_dtype() -> jnp.dtype:
    """Returns the default CRIKit JAX dtype

    :returns: CRIKit default JAX dtyle
    :rtype: jnp.dtype

    """
    return _default_dtype


# similar to what pyadjoint defines in adjfloat.py
def annotate_operator(orig_operator, nojit=False):
    """Decorates an operator like __add__, __sub__, etc.
    with JAX JIT compilation.
    """

    def reverse(block):
        def _reversed_block(operator, args, output):
            return block(operator, (args[1], args[0]), output)

        return _reversed_block

    op_map = {
        "__neg__": (lambda x: -x, NegBlock),
        "__add__": (lambda x, y: x + y, AddBlock),
        "__mul__": (lambda x, y: x * y, MulBlock),
        "__truediv__": (lambda x, y: x / y, DivBlock),
        "__rtruediv__": (lambda x, y: y / x, reverse(DivBlock)),
        "__sub__": (lambda x, y: x - y, SubBlock),
        "__pow__": (lambda x, y: x ** y, PowBlock),
        "__radd__": (lambda x, y: y + x, reverse(AddBlock)),
        "__rmul__": (lambda x, y: y * x, reverse(MulBlock)),
        "__rsub__": (lambda x, y: y - x, reverse(SubBlock)),
        "__rpow__": (lambda x, y: y ** x, reverse(PowBlock)),
    }

    def is_unitary(op):
        # only one unary operator right now
        return op.__name__ == "__neg__"

    op, block_ctor = op_map[orig_operator.__name__]
    if not nojit:
        op = jax.jit(op)

    def annotated_operator(*args):
        # args[0] is always self
        output = args[0].__class__(op(*(convert_arg(x) for x in args)))
        args = [
            arg if isinstance(arg, OverloadedType) else create_overloaded_object(arg)
            for arg in args
        ]
        if annotate_tape():
            block = block_ctor(op, args, output)
            tape = get_working_tape()
            tape.add_block(block)
            block.add_output(output.block_variable)

        return output

    annotated_operator.__name__ = orig_operator.__name__
    return annotated_operator


@register_pytree_node_class
class ndarray(OverloadedType):
    def __repr__(self):
        return f"ndarray({self.arr.__repr__()})"

    def __array__(self, dtype=None):
        if dtype:
            if hasattr(self.arr, "__array__"):
                return self.arr.__array__(dtype=dtype)
            else:
                return np.array(self.arr).__array__(dtype=dtype)
        else:
            if hasattr(self.arr, "__array__"):
                return self.arr.__array__()
            else:
                return np.array(self.arr).__array__()

    def __float__(self):
        return float(self.unwrap(True))

    def __iter__(self):
        return self.arr.__iter__()

    def __init__(self, obj: Array, *args, **kwargs):
        """Note: you should not typically use this constructor directly in your code. Instead, you should call
        :func:`array` or :func:`asarray`, which will call the constructor of this class when appropriate.


        :param obj: the object to wrap; should be either a JAX ndarray or something
            that can be converted to one (such as a list or tuple of floats or ints,
            or a float or int, or a numpy ndarray)
        :type obj: jax.interpreters.xla.DeviceArray
        :return: a class that wraps a JAX array (such that it can be added to the JAX Pytree
            and thus used as an argument to a differentiable function) to be passed to a
            function wrapped with `overload_jax()`, while also inheriting from
            pyadjoint.OverloadedType
            (since you can't inherit from a JAX array;
            see https://github.com/google/jax/issues/4269).
        :rtype: ndarray

        """
        dtype = kwargs.get("dtype", np.float64)
        self.arr = obj
        self.extras = args
        super().__init__()

    @property
    def value(self):
        return self.arr

    @value.setter
    def value(self, new_jax_array):
        self.arr = new_jax_array

    @property
    def dtype(self):
        return self.arr.dtype

    @property
    def ndim(self):
        return self.arr.ndim

    def unwrap(self, to_jax: bool = True) -> jnp.ndarray:
        """
        If this ndarray holds recursively nested ndarrays (e.g. its __repr__() is ndarray(ndarray(...))), unwrap until it holds the array data contained in the deepest-nested ndarray.

        This is mostly a utility for use in jacfwd and jacrev in pyadjoint_utils/numpy_adjoint/jax.py

        :param to_jax: go one level further and return the raw JAX array (instead of ndarray, the OverloadedType wrapper)?, defaults to False
        :type to_jax: bool, optional
        :return: unwrapped version of self
        :rtype: jax.interpreters.xla.DeviceArray
        """
        newarr = self.arr
        while isinstance(newarr, ndarray):
            newarr = newarr.arr

        if isinstance(newarr, (list, tuple)):
            tna = type(newarr)
            arr = list(newarr)
            for i, val in enumerate(newarr):
                if isinstance(val, ndarray):
                    val = val.unwrap(to_jax=to_jax)
                arr[i] = val
            return tna(arr)

        if to_jax:
            return newarr
        else:
            # we went one level too far!
            # whatever, it's relatively cheap to construct a new ndarray
            return ndarray(newarr)

    def tree_flatten(self) -> Tuple[Tuple[jnp.ndarray, ...], None]:
        """
        Flattens an ndarray in the JAX Pytree structure

        :return:  tuple containing any arrays (or other children) this ndarray holds, and an empty metadata field
        :rtype: tuple

        """
        if self.extras:
            return ((self.arr, *self.extras), None)
        return ((self.arr,), None)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Constructs an ndarray from its flattened components

        :param cls: ndarray
        :type cls: type
        :param aux_data: ignore this parameter
        :type aux_data: None
        :param children: any children that belonged to this ndarray before it was flattened
        :type children: Union[jax.interpreters.xla.DeviceArray,Iterable[jax.interpreters.xla.DeviceArray]]
        :return: an ndarray holding the children
        :rtype: ndarray

        """
        return cls(*children)

    def flatten(self) -> jnp.ndarray:
        """Returns a flattened 1-d array (NOT an ndarray, but rather the array type it contains)"""
        try:
            return self.arr.flatten()
        except Exception:
            return self.arr

    @classmethod
    def _ad_init_object(cls, obj: Array):
        obj = jnp.array(obj) if not isinstance(obj, (int, float)) else obj
        return cls(obj)

    @property
    def size(self):
        """How many elements does the array contain?"""
        try:
            return self.arr.size
        except Exception:
            return 0

    @property
    def shape(self):
        """The shape of the array"""
        try:
            return self.arr.shape
        except Exception:
            return ()

    @property
    def T(self):
        """Returns the transpose of this ndarray"""
        return ndarray(self.arr.T)

    def __len__(self):
        return self.arr.__len__()

    def __eq__(self, other):
        return self.arr.__eq__(other)

    # annotated operators are implemented in the table
    # mapping their name to a tuple of the implementation and
    # the corresponding Block contained in annotate_operator()
    @annotate_operator
    def __add__(self, other):
        pass

    @annotate_operator
    def __neg__(self):
        pass

    @annotate_operator
    def __truediv__(self, other):
        pass

    @annotate_operator
    def __rtruediv__(self, other):
        pass

    @annotate_operator
    def __radd__(self, other):
        return self.__add__(other)

    @annotate_operator
    def __rmul__(self, other):
        pass

    def __iadd__(self, other):
        return NotImplemented

    def __imul__(self, other):
        return NotImplemented

    def __isub__(self, other):
        return NotImplemented

    @annotate_operator
    def __sub__(self, other):
        pass

    @annotate_operator
    def __rsub__(self, other):
        pass

    @annotate_operator
    def __mul__(self, other):
        pass

    @annotate_operator
    def __pow__(self, other):
        pass

    @annotate_operator
    def __rpow__(self, other):
        pass

    def __abs__(self):
        return ndarray(jnp.abs(self.arr))

    def _ad_create_checkpoint(self) -> jnp.ndarray:
        return self.arr

    def _ad_restore_at_checkpoint(self, checkpoint):
        return ndarray(checkpoint)

    def _ad_dim(self) -> int:
        return self.arr.size

    def _ad_dot(self, other) -> float:
        sflat = self.flat()
        oflat = flatten(other)
        if sflat.size == 1 or oflat.size == 1:
            return float(jnp.sum(sflat * oflat))
        return float(jnp.dot(sflat, oflat))

    def _ad_mul(self, other):
        return ndarray(self.arr * other)

    def _ad_add(self, other):
        return ndarray(self.arr + other)

    def _ad_copy(self):
        if isinstance(self.arr, tuple):
            try:
                return ndarray(tuple(map(lambda x: x.copy(), self.arr)))
            except Exception:
                return self.arr
        return (
            ndarray(self.arr.copy())
            if not isinstance(self.arr, (int, float))
            else ndarray(self.arr)
        )

    def _ad_convert_type(self, value, options={}):
        return array(value, **options)

    def copy(self, *args, **kwargs):
        return ndarray(self.arr.copy(*args, **kwargs))

    def copy_data(self):
        return ndarray(self.arr.copy())

    def flat(self) -> jnp.ndarray:
        return jnp.ravel(self.unwrap(to_jax=True))

    @staticmethod
    def _ad_assign_numpy(dst, src, offset):
        if isinstance(src, ndarray):
            src = src.unwrap(to_jax=True)
        if isinstance(src, (list, tuple)):
            src = list(src)
            while isinstance(src, list) and len(src) == 1:
                src = src[0]
            else:
                for i, s in enumerate(src):
                    if isinstance(s, ndarray):
                        src[i] = s.unwrap(to_jax=True)
                src_val = src[offset]
                if isinstance(src_val, (int, float)):
                    dst = src_val
                else:
                    dst = jnp.reshape(jnp.array(src_val), dst.shape)
                offset += 1
                return array(dst), offset

        if hasattr(src, "__len__") and len(src) == 1:
            src = src[0]
        if isinstance(src, (int, float)):
            return array(src), offset + 1
        dst = jnp.reshape(jnp.array(src[offset : offset + dst.size]), dst.shape)
        offset += dst.size
        return array(dst), offset

    @staticmethod
    def _ad_to_list(m) -> list:
        if isinstance(m, ndarray):
            return np.array(m.arr).flatten().tolist()
        try:
            return list(np.array(m).tolist())
        except TypeError:
            return [m]

    def __getitem__(self, item):
        annotate = annotate_tape()
        if annotate:
            block = JAXArraySliceBlock(self, item)
            tape = get_working_tape()
            tape.add_block(block)

        with stop_annotating():
            out = self.arr.__getitem__(item)

        if annotate:
            out = ndarray(out) if not isinstance(out, ndarray) else out
            block.add_output(out.create_block_variable())
        return out


def array(obj: Array, **kwargs) -> ndarray:
    """Converts the input to an :class:`ndarray`. This function is NOT
        overloaded (does not add any :class:`Block` to the :class:`Tape`).
        If you want to convert an :class:`AdjFloat` to an :class:`ndarray`
        or vice-versa, use the functions :func:`to_jax` or :func:`to_adjfloat`
        respectively.

    :param obj: the object to wrap; should be either a JAX ndarray or something
        that can be converted to one (such as a list or tuple of floats or ints,
        or a float or int, or a numpy ndarray)
    :type obj: Union[jax.interpreters.xla.DeviceArray,Iterable[Union[int,float,jax.interpreters.xla.DeviceArray]]]
    :returns: a class that wraps a JAX array (such that it can be added to the JAX Pytree
        and thus used as an argument to a differentiable function) to be passed to a
        function wrapped with `overload_jax()`, while also inheriting from
        pyadjoint.OverloadedType
        (since you can't inherit from a JAX array;
        see https://github.com/google/jax/issues/4269).
    :rtype: ndarray

    """
    return _backend_array(obj, **kwargs)


def _backend_array(obj, **kwargs):
    directly_convertible_types = (
        float,
        int,
        jax.interpreters.xla.DeviceArray,
        jnp.ndarray,
        np.ndarray,
        ndarray,
    )
    if isinstance(obj, ndarray):
        return obj
    elif isinstance(obj, directly_convertible_types):
        return ndarray(obj)

    dtype = kwargs.get("dtype", _default_dtype)

    dtype = kwargs.get("dtype", _default_dtype)
    if isinstance(obj, (list, tuple)):
        if len(obj) == 0:
            return ndarray(obj)
        while isinstance(obj, (list, tuple)) and not isinstance(
            obj[0], directly_convertible_types
        ):
            if len(obj) == 1:
                obj = obj[0]
            else:
                obj = list(obj)
                for i, val in enumerate(obj):
                    obj[i] = array(val)

        if isinstance(obj[0], ndarray):
            return ndarray(obj)

    elif isinstance(obj, AdjFloat):
        obj = float(obj)
    elif isinstance(obj, Control):
        obj = obj.data()

    # jax.numpy.array(numpy.array(x)) is typically faster than
    # jax.numpy.array(x), especially if x is a list or tuple, because
    # numpy.array() is written in C, whereas jax.numpy.array() is written in
    # Python. This is recommended by the JAX developers.
    try:
        return ndarray(jnp.array(np.array(obj), dtype=dtype), **kwargs)
    except Exception:
        return ndarray(obj)


def asarray(obj: Array, **kwargs) -> ndarray:
    return array(obj, **kwargs)


ConvertJAXToAdjFloat = make_convert_block(
    AdjFloat,
    _backend_array,
    "ConvertJAXToFloat",
)

ConvertAdjFloatToJAX = make_convert_block(
    _backend_array,
    AdjFloat,
    "ConvertAdjFloatToJAX",
)


to_adjfloat = overload_function(AdjFloat, ConvertJAXToAdjFloat)

to_jax = overload_function(_backend_array, ConvertAdjFloatToJAX)


def flatten(x):
    if isinstance(x, ndarray):
        return x.flat()
    elif isinstance(x, (np.ndarray, jax.interpreters.xla.DeviceArray)):
        return jnp.ravel(x)
    else:
        return x


class JAXArraySliceBlock(Block):
    def __init__(self, arr, item):
        super().__init__()
        self.add_dependency(arr)
        self.item = item

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        adj_output = jnp.zeros(inputs[0].shape)
        return index_update(adj_output, self.item, adj_inputs[0])

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return inputs[0][self.item]

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return tlm_inputs[0][self.item]


def convert_arg(
    x: Union[ndarray, list, tuple, int, float, jnp.ndarray]
) -> Union[list, tuple, int, float, jnp.ndarray]:
    if isinstance(x, (jnp.ndarray, jax.interpreters.xla.DeviceArray, int, float)):
        return x
    elif isinstance(x, ndarray):
        return x.unwrap(to_jax=True)

    elif isinstance(x, (list, tuple)):
        tx = type(x)
        return tx([convert_arg(v) for v in x])
    elif isinstance(x, Control):
        return convert_arg(x.data())
    try:
        return jnp.array(x)
    except Exception:
        return x


register_overloaded_type(ndarray, jnp.ndarray)
register_overloaded_type(ndarray, jax.interpreters.xla.DeviceArray)
try:
    register_overloaded_type(ndarray, jax.interpreters.xla._DeviceArray)
except:
    pass

try:
    import jaxlib

    register_overloaded_type(ndarray, jaxlib.xla_extension.Buffer)
except:
    pass
